      SUBROUTINE PDLAEVSWP( N, ZIN, LDZI, ZOUT, IZ, JZ, DESCZ, NVS, KEY,
     $                      WORK, LWORK )
*
*
*
*  -- ScaLAPACK routine (version 1.0) --
*     University of Tennessee, Knoxville, Oak Ridge National Laboratory,
*     and University of California, Berkeley.
*     February 28, 1995
*
*     .. Scalar Arguments ..
      INTEGER            IZ, JZ, LDZI, LWORK, N
*     ..
*     .. Array Arguments ..
      INTEGER            DESCZ( 8 ), KEY( * ), NVS( * )
      DOUBLE PRECISION   WORK( * ), ZIN( LDZI, * ), ZOUT( * )
*     ..
*
*  Purpose
*  =======
*
*  PDLAEVSWP moves the eigenvectors from where they are computed,
*  to a ScaLAPACK standard block cyclic array.
*
*  Notes
*  =====
*
*
*  A description vector is associated with each 2D block-cyclicly dis-
*  tributed matrix.  This vector stores the information required to
*  establish the mapping between a matrix entry and its corresponding
*  process and memory location.
*
*  In the following comments, the character _ should be read as
*  "of the distributed matrix".  Let A be a generic term for any 2D
*  block cyclicly distributed matrix.  Its description vector is DESCA:
*
*  NOTATION        STORED IN  EXPLANATION
*  --------------- ---------- ------------------------------------------
*  M_A    (global) DESCA( 1 ) The number of rows in the distributed
*                             matrix.
*  N_A    (global) DESCA( 2 ) The number of columns in the distributed
*                             matrix.
*  MB_A   (global) DESCA( 3 ) The blocking factor used to distribute
*                             the rows of the matrix.
*  NB_A   (global) DESCA( 4 ) The blocking factor used to distribute
*                             the columns of the matrix.
*  RSRC_A (global) DESCA( 5 ) The process row over which the first row
*                             of the matrix is distributed.
*  CSRC_A (global) DESCA( 6 ) The process column over which the first
*                             column of the matrix is distributed.
*  CTXT_A (global) DESCA( 7 ) The BLACS context handle, indicating the
*                             global context of the operation on the
*                             matrix. The context itself is global.
*  LLD_A  (local)  DESCA( 8 ) The leading dimension of the local array
*                             storing the local blocks of the distri-
*                             buted matrix A. LLD_A >= MAX(1,LOCp(M_A)).
*
*  Let K be the number of rows or columns of a distributed matrix,
*  and assume that its process grid has dimension p x q.
*  LOCp( K ) denotes the number of elements of K that a process
*  would receive if K were distributed over the p processes of its
*  process column.
*  Similarly, LOCq( K ) denotes the number of elements of K that a
*  process would receive if K were distributed over the q processes of
*  its process row.
*  The values of LOCp() and LOCq() may be determined via a call to the
*  ScaLAPACK tool function, NUMROC:
*          LOCp( M ) = NUMROC( M, MB_A, MYROW, RSRC_A, NPROW ),
*          LOCq( N ) = NUMROC( N, NB_A, MYCOL, CSRC_A, NPCOL ).
*
*  Arguments
*  =========
*
*     NP = the number of rows local to a given process.
*     NQ = the number of columns local to a given process.
*
*  N       (global input) INTEGER
*          The order of the matrix A.  N >= 0.
*
*  ZIN     (local input) DOUBLE PRECISION array,
*          dimension ( LDZI, NVS(iam) )
*          The eigenvectors on input.  Each eigenvector resides entirely
*          one process.  Each process holds a contiguous set of
*          NVS(iam) eigenvectors.  The first eigenvector which the
*          process holds is:  sum for i=[0,iam-1) of NVS(i)
*
*  LDZI    (locl input) INTEGER
*          leading dimension of the ZIN array
*
*  ZOUT    (local output) DOUBLE PRECISION array
*          global dimension (N, N), local dimension (DESCZ(8), NQ)
*          The eigenvectors on output.  The eigenvectors are distributed
*          in a block cyclic manner in both dimensions, with a
*          block size of NB.
*
*
*  IZOUT      (global input) INTEGER
*          The global row index of the submatrix of the distributed
*          matrix ZOUT to operate on.
*
*
*  JZOUT      (global input) INTEGER
*          The global column index of the submatrix of the distributed
*          matrix ZOUT to operate on.
*
*
*  DESCZOUT   (global and local input) INTEGER array of dimension 8
*          The array descriptor for the distributed matrix ZOUT.
*
*  NVS     (global input) INTEGER array, dimension( nprocs+1 )
*          nvs(i) = number of processes
*          number of eigenvectors held by processes [0,i-1)
*          nvs(1) = number of eigen vectors held by [0,1-1) == 0
*          nvs(nprocs+1) = number of eigen vectors held by [0,nprocs) ==
*            total number of eigenvectors
*
*  KEY     (global input) INTEGER array, dimension( N )
*          Indicates the actual index (after sorting) for each of the
*          eigenvectors.
*
*  WORK    (local workspace) double precision array, dimension (LWORK)
*
*  LWORK   (local input) INTEGER dimension of WORK
*
*     .. Local Scalars ..
      INTEGER            CYCLIC_I, CYCLIC_J, DIST, I, IAM, II, INCII, J,
     $                   MAXI, MAXII, MINI, MINII, MYCOL, MYROW, NB,
     $                   NBUFSIZE, NPCOL, NPROCS, NPROW, PCOL, RECVCOL,
     $                   RECVFROM, RECVROW, SENDCOL, SENDROW, SENDTO
*     ..
*     .. External Functions ..
      INTEGER            INDXG2L, INDXG2P
      EXTERNAL           INDXG2L, INDXG2P
*     ..
*     .. External Subroutines ..
      EXTERNAL           BLACS_GRIDINFO, DGERV2D, DGESD2D
*     ..
*     .. Intrinsic Functions ..
      INTRINSIC          MAX, MIN, MOD
*     ..
*     .. Executable Statements ..
      CALL BLACS_GRIDINFO( DESCZ( 7 ), NPROW, NPCOL, MYROW, MYCOL )
      IAM = MYROW + MYCOL*NPROW
      IAM = MYROW*NPCOL + MYCOL
*
      NB = DESCZ( 3 )
*
      NPROCS = NPROW*NPCOL
*
      DO 100 DIST = 0, NPROCS - 1
*
         SENDTO = MOD( IAM+DIST, NPROCS )
         RECVFROM = MOD( NPROCS+IAM-DIST, NPROCS )
*
         SENDROW = MOD( SENDTO, NPROW )
         SENDCOL = SENDTO / NPROW
         RECVROW = MOD( RECVFROM, NPROW )
         RECVCOL = RECVFROM / NPROW
*
         SENDROW = SENDTO / NPCOL
         SENDCOL = MOD( SENDTO, NPCOL )
         RECVROW = RECVFROM / NPCOL
         RECVCOL = MOD( RECVFROM, NPCOL )
*
*        Figure out what I have that process "sendto" wants
*
         NBUFSIZE = 0
*
*        We are looping through the eigenvectors that I presently own.
*
         DO 30 J = NVS( 1+IAM ) + JZ, NVS( 1+IAM+1 ) + JZ - 1
            PCOL = INDXG2P( KEY( J ), DESCZ( 4 ), -1, DESCZ( 6 ),
     $             NPCOL )
            IF( SENDCOL.EQ.PCOL ) THEN
               MINII = MOD( SENDROW+DESCZ( 5 ), NPROW )*DESCZ( 3 ) + 1
               MAXII = DESCZ( 1 )
               INCII = DESCZ( 3 )*NPROW
               DO 20 II = MINII, MAXII, INCII
                  MINI = MAX( II, IZ )
                  MAXI = MIN( II+DESCZ( 3 )-1, N+IZ-1 )
                  DO 10 I = MINI, MAXI, 1
                     NBUFSIZE = NBUFSIZE + 1
                     WORK( NBUFSIZE ) = ZIN( I+1-IZ,
     $                                  J-NVS( 1+IAM )+1-JZ )
   10             CONTINUE
   20          CONTINUE
            END IF
   30    CONTINUE
*
*
         IF( MYROW.NE.SENDROW .OR. MYCOL.NE.SENDCOL )
     $      CALL DGESD2D( DESCZ( 7 ), NBUFSIZE, 1, WORK, NBUFSIZE,
     $                    SENDROW, SENDCOL )
*
*
*        Figure out what process "recvfrom" has that I want
*
         NBUFSIZE = 0
         DO 60 J = NVS( 1+RECVFROM ) + JZ,
     $           NVS( 1+RECVFROM+1 ) + JZ - 1, 1
            PCOL = INDXG2P( KEY( J ), DESCZ( 4 ), -1, DESCZ( 6 ),
     $             NPCOL )
            IF( MYCOL.EQ.PCOL ) THEN
               MINII = MOD( MYROW+DESCZ( 5 ), NPROW )*DESCZ( 3 ) + 1
               MAXII = DESCZ( 1 )
               INCII = DESCZ( 3 )*NPROW
               DO 50 II = MINII, MAXII, INCII
                  MINI = MAX( II, IZ )
                  MAXI = MIN( II+NB-1, N+IZ-1 )
                  DO 40 I = MINI, MAXI, 1
                     NBUFSIZE = NBUFSIZE + 1
   40             CONTINUE
   50          CONTINUE
            END IF
   60    CONTINUE
*
*
*
         IF( MYROW.NE.RECVROW .OR. MYCOL.NE.RECVCOL )
     $      CALL DGERV2D( DESCZ( 7 ), 1, NBUFSIZE, WORK, 1, RECVROW,
     $                    RECVCOL )
*
         NBUFSIZE = 0
         DO 90 J = NVS( 1+RECVFROM ) + JZ,
     $           NVS( 1+RECVFROM+1 ) + JZ - 1, 1
            PCOL = INDXG2P( KEY( J ), DESCZ( 4 ), -1, DESCZ( 6 ),
     $             NPCOL )
            IF( MYCOL.EQ.PCOL ) THEN
               CYCLIC_J = INDXG2L( KEY( J ), DESCZ( 3 ), -1, -1, NPCOL )
               CYCLIC_I = 1
               MINII = MOD( MYROW+DESCZ( 5 ), NPROW )*DESCZ( 3 ) + 1
               MAXII = DESCZ( 1 )
               INCII = DESCZ( 3 )*NPROW
               DO 80 II = MINII, MAXII, INCII
                  MINI = MAX( II, IZ )
                  CYCLIC_I = INDXG2L( MINI, DESCZ( 3 ), -1, -1, NPROW )
                  MAXI = MIN( II+NB-1, N+IZ-1 )
                  DO 70 I = MINI, MAXI, 1
                     NBUFSIZE = NBUFSIZE + 1
                     ZOUT( CYCLIC_I+( CYCLIC_J-1 )*DESCZ( 8 ) )
     $                  = WORK( NBUFSIZE )
                     CYCLIC_I = CYCLIC_I + 1
   70             CONTINUE
   80          CONTINUE
            END IF
   90    CONTINUE
*
  100 CONTINUE
      RETURN
      END
